"""

test script added for testing sub documents as they might overwrite each other and we
want this to be prevented. This is taken from camper so hence the names

"""

from mongogogo import *
import datetime

class Location(Schema):
    name = String()

class BarcampSchema(Schema):
    name                = String()
    location            = Location()
    locations           = List(Location())

class Barcamp(Record):

    schema = BarcampSchema()
    default_values = {
        'location'      : {}, # this might be a problem due to being an object
        'locations'     : [], # this might be a problem due to being an object
    }

class Barcamps(Collection):
    
    data_class = Barcamp

def test_subdict_with_default(db):
    """initially we had a bug where the dict in default_values was overwritten all the time
    because we did not copy it, yielding strange results in batch operations. This test makes sure 
    we don't run into it again for sub documents
    """
    barcamps = Barcamps(db.barcamps)
    for i in range(1,4):
        data = dict(
            name = str(i),
            location = {
                'name' : str(i)
            },
            locations = [{'name' : str(i)}]
        )
        barcamp = Barcamp(data, collection = barcamps)
        barcamp.put()

    # also in a loop we should keep the correct location information
    bcs = barcamps.find()
    a = []
    for b in bcs:
        a.append(b)
    for b in a:
        # all names should be equal here as this is how we set this up
        assert b['name'] == b.location['name']

def test_lists_with_default(db):
    """we test the same here for the list default"""
    barcamps = Barcamps(db.barcamps)
    for i in range(1,4):
        data = dict(
            name = str(i),
            locations = [{'name' : str(i)}]
        )
        barcamp = Barcamp(data, collection = barcamps)
        barcamp.put()

    # also in a loop we should keep the correct location information
    bcs = barcamps.find()
    a = []
    for b in bcs:
        a.append(b)
    for b in a:
        assert b['locations'][0]['name'] == b['name']



import pymongo
import datetime
from mongogogo import Record, Collection
from mongogogo import String, DateTime, Integer, Dict, Default
from mongogogo import SchemaNode, Schema

DB_NAME = "mongogogo_testing_78827628762"

def setup_db():
    db = pymongo.Connection()[DB_NAME]
    return db

def teardown_db(db):
    #pymongo.Connection().drop_database(DB_NAME)
    db.persons.remove()
    db.barcamps.remove()

def pytest_funcarg__db(request):
    return request.cached_setup(
        setup = setup_db,
        teardown = teardown_db,
        scope = "function")

class Incrementor(SchemaNode):  

    def do_deserialize(self, value, data, **kw):
        return int(value)+1

class BioType(Schema):
    name = String()
    url = String()

class PersonSchema(Schema):
    firstname = String(required=True)
    lastname = String(default="foobar", required=True)
    creation = DateTime(required=True, default = datetime.datetime.utcnow)
    age = Integer(default=24)
    incr = Incrementor(default=1)
    d = Dict(default={}, dotted=True)
    e = Dict(default={}, dotted=False)
    #bio = BioType(on_serialize=[Default({'name': 'foobar'})])

class Person(Record):
    schema = PersonSchema()
    default_values = {
        'lastname' : 'foobar',
        'creation' : datetime.datetime.utcnow,
        'age'      : 24,
        'incrementor'      : 1,
        'd'      : {},
        'e'      : {},
        'bio'    : {'name' : 'foobar'},
    }

class Persons(Collection):
    data_class = Person

class SchemalessPerson(Record):
    schema = PersonSchema()
    schemaless = True

class SchemalessPersons(Collection):
    data_class = SchemalessPerson

def pytest_funcarg__persons(request):
    db = request.getfuncargvalue("db")
    return Persons(db.persons)

def pytest_funcarg__schemaless_persons(request):
    db = request.getfuncargvalue("db")
    return SchemalessPersons(db.persons)



from utils import Invalid, null

class Filter(object):
    """A Filter can be used on serialization and deserialization. It can either convert an incoming value to something
    else but also can do validation. In case it fails it's supposed to raise an ``Invalid`` exception."""

    def __call__(self, value, data, **kw):
        """call the actual filter. Passed in is the value to process in ``value`` and the complete incoming record
        in ``data``. Additional keyword arguments passed to ``serialize()`` or ``deserialize()`` are also passed into
        each filter. That way you can e.g. pass in database connections or similar constructs you might need inside
        a filter."""

        return value


class Default(Filter):
    """a filter which sets a default value if it is missing. Beware to not set this if you set required to True."""

    def __init__(self, item):
        """initialize the Default filter with the value to set. If ``item`` is callable then it will be called
        and the result will be set as default value"""
        self.item = item

    def __call__(self, value, data, **kw):
        """process the value"""
        if value is null:
            if callable(self.item):
                return self.item()
            else:
                return self.item
        return value

class EqualTo(Filter):
    """a filter which sets a default value if it is missing. Beware to not set this if you set required to True"""

    def __init__(self, field):
        self.field  = field

    def __call__(self, value, data, coll = None, **kw):
        """process the value"""
        if value != data[self.field]:
            raise Invalid(self, "fields do not match")
        return value




from mongogogo.schema import null, Invalid, String
import pytest

def test_basic_string_serialize_ok(schema1):
    s = String()
    v = s.serialize("foobar")
    assert v == "foobar"

def test_basic_string_serialize_not_required_and_missing(schema1):
    s = String()
    v = s.serialize()
    assert v == null

def test_basic_string_serialize_required_and_missing(schema1):
    s = String(required = True)
    pytest.raises(Invalid, s.serialize)

def test_basic_string_serialize_default_and_missing(schema1):
    s = String(required = True, default="foobar")
    v = s.serialize()
    assert v == "foobar"

def test_basic_string_serialize_default_and_missing(schema1):
    s = String(required = True, default="foobar")
    v = s.serialize()
    assert v == "foobar"


def test_basic_string_serialize_default_and_missing(schema1):
    s = String(required = True, default="foobar")
    v = s.serialize()
    assert v == "foobar"


def test_basic_string_set_to_none(schema1):
    s = String()
    v = s.serialize(None)
    assert v is None

def test_string_too_long(schema1):
    s = String(max_length=10)
    s.serialize("0123456789")
    pytest.raises(Invalid, s.serialize, '01234567890')
    


import types

class Null(object):
    """a null object. We use a separate class here so that we can control the representation"""

    def __init__(self, label = "<null>"):
        """initialize the null object with a label describing it's use"""
        self.label = label

    def __repr__(self):
        return self.label

null = None
#null = Null("<null>")
marker = Null("<marker>")


class Invalid(Exception):
    """exception which is called whenever serialization or deserialization has not worked"""

    def __init__(self, node, msg=""):
        """something went wrong when serializing or deserializing something"""
        self.node = node
        self.msg = msg

    def __str__(self):
        """return something printable"""
        return "<%s: %s in %s>" %(self.__class__.__name__, self.msg, self.node.name)


class AttributeMapper(dict):
    """a dictionary like object which also is accessible via getattr/setattr"""

    __slots__ = []

    def __init__(self, default={}, *args, **kwargs):
        super(AttributeMapper, self).__init__(*args, **kwargs)
        self.update(default)
        self.update(kwargs)

    def __getattr__(self, k):
        """retrieve some data from the dict"""
        if self.has_key(k):
            return self[k]
        raise AttributeError(k)

    def __setattr__(self, k,v):
        """store an attribute in the map"""
        self[k] = v

    def _clone(self):
        """return a clone of this object"""
        d = copy.deepcopy(self) 
        return AttributeMapper(d)

    def update(self, d):
        """update the dictionary but make sure that existing included AttributeMappers are only updated aswell"""
        for a,v in d.items():
            if a not in self:
                self[a] = v
            elif isinstance(self[a], AttributeMapper) and type(v) == types.DictType:
                self[a].update(v)
            elif type(self[a]) == types.DictType and type(v) == types.DictType:
                self[a].update(v)
            else:
                self[a] = v



from mongogogo.schema import null, Invalid
import pytest

def test_basic_serialize_ok(schema1):
    
    data = {'not_required' : 'n/a',
         'required' : 'Required',
         'with_default' : 'no default'}
    res = schema1.serialize(data)
    assert res['not_required'] == "n/a"
    assert res['required'] == "Required"
    assert res['with_default'] == "no default"

def test_basic_serialize_with_default(schema1):
    
    data = {'not_required' : 'n/a',
         'required' : 'Required'}
    res = schema1.serialize(data)
    assert res['not_required'] == "n/a"
    assert res['required'] == "Required"
    assert res['with_default'] == "default"

def test_basic_serialize_with_missing(schema1):
    
    data = {'required' : 'Required'}
    res = schema1.serialize(data)
    assert res['not_required'] is null
    assert res['required'] == "Required"
    assert res['with_default'] == "default"

def test_basic_serialize_not_required(schema1):
    
    data = {'not_required' : 'Required'}
    pytest.raises(Invalid, schema1.serialize, data)

def test_basic_serialize_submapping_ok(schema2):
    data = {
        'bio1' : {
            'name' : 'Foo', 
            'url' : 'http://example.com', 
        },
        'bio2' : {
            'name' : 'Foo', 
            'url' : 'http://example.com', 
        }
    }
    res = schema2.serialize(data)
    assert res['bio1']['name'] == "Foo"
    assert res['bio2']['name'] == "Foo"
    assert res['bio1']['url'] == "http://example.com"
    assert res['bio2']['url'] == "http://example.com"

            
def test_basic_serialize_submapping_missing_bio2(schema2):
    data = {
        'bio1' : {
            'name' : 'Foo', 
            'url' : 'http://example.com', 
        },
    }
    pytest.raises(Invalid, schema2.serialize, data)

def test_basic_serialize_submapping_missing_bio1(schema2):

    data = {
        'bio2' : {
            'name' : 'Foo', 
            'url' : 'http://example.com', 
        },
    }

    res = schema2.serialize(data)
    assert res['bio1']['name'] == "foobar"
    assert res['bio2']['name'] == "Foo"
    assert res['bio1']['url'] == null
    assert res['bio2']['url'] == "http://example.com"
            

def test_basic_serialize_with_subclass(subschema1):
    
    data = {'not_required' : 'n/a',
         'name2' : 'hansi', # new field in this subschema
         'required' : 'Required',
         'with_default' : 'no default'}
    res = subschema1.serialize(data)
    assert res['not_required'] == "n/a"
    assert res['required'] == "Required"
    assert res['with_default'] == "no default"
    assert res['name2'] == "hansi"


def test_underscore_mapping(underscoreschema):
    
    data = {'_name' : 'foobar'}
    res = underscoreschema.serialize(data)
    assert res['_name'] == "foobar"



from mongogogo.schema import *
from conftest import TestSchema1
import pytest

class MyDict(dict):
    """custom dict for testing"""
    myname = "NameSchema"
    CUSTOM = True

def test_deserialize_to_class_simple():
    schema1 = TestSchema1(kls=MyDict)
    data = {'not_required' : 'n/a',
         'required' : 'Required',
         'with_default' : 'no default'}
    res = schema1.serialize(data)
    res = schema1.deserialize(res)
    assert res.CUSTOM == True



##############################################


def test_deserialize_to_class_list(schema1):

    class NameSchema(Schema):
        name = String()

    class Name(Schema):
        """content class for name schema"""

    class NamesSchema(Schema):
        """schema containing a list of names"""
        names = List(NameSchema(kls=Name)) 

    class Names(dict):
        """Content class for NamesSchema""" 
        myname = "NamesSchema"

    names = NamesSchema(kls=Names)
    data = {'names' : [{'name' : 'one'}, {'name' : 'two'}, {'name' : 'three'}]}
    res = names.serialize(data)
    res = names.deserialize(res)
    assert res.__class__.__name__ == "Names"

    # now check children
    assert res['names'][0].__class__.__name__ == "Name"




def test_fillin_defaults_in_subsubschema():

    class RecordSchema(Schema):
        test = String()
        empty = String(default = "")
        b = Boolean(default = False)
        foo = String(default = "bar")

    class SubDictSchema(Schema):
        test_list = List(RecordSchema())

    class TestSchema(Schema):

        sub1 = SubDictSchema(default = {
                'test_list' : []
            })

    schema = TestSchema()
    payload = {
        'sub1' : {
            'test_list' : 
                [
                    {'test' : 'testing'}
                ]
        } # sub1
    } # payload

    data = schema.serialize(payload)

    assert data['sub1']['test_list'][0]['test'] == 'testing'
    assert data['sub1']['test_list'][0]['foo'] == 'bar'
    assert data['sub1']['test_list'][0]['b'] == False
    assert data['sub1']['test_list'][0]['empty'] == ""





from mongogogo.schema.filters import *
from mongogogo.schema.nodes import *

class SomeType(Schema):
    
    url = String()
    description = String()

class Bio(Schema):
    name = String()
    url = String()

class TestSchema1(Schema):
    """very simply schema to get started"""
    not_required = String()
    required = String(required=True)
    with_default = String(on_serialize=[Default("default")])

class TestSchema2(Schema):
    """very simply schema with a sub schema"""
    name = String()
    bio1 = Bio(on_serialize=[Default({'name': 'foobar'})])
    bio2 = Bio(required = True)

    
class UnderscoreTestSchema(Schema):
    """very simply schema with a sub schema"""
    _name = String()

class TestSchema12(Schema):
    name = String()
    permissions = List(String(), on_serialize = [Default([])])
    links = List(SomeType())
    bio = Bio(on_serialize=[Default({'name': 'foobar'})])

class SubSchema(TestSchema1):
    name2 = String()

def pytest_funcarg__schema1(request):
    return TestSchema1()

def pytest_funcarg__schema2(request):
    return TestSchema2()

def pytest_funcarg__subschema1(request):
    return SubSchema()

def pytest_funcarg__underscoreschema(request):
    return UnderscoreTestSchema()

